"""
Provides a FileCache and LRUFileCache using only file-based synchronization
primitives.
"""
import os
import time
import stat
import errno
from math import ceil
from pprint import pformat
import md5 # use hashlib
from tempfile import gettempdir
import lib.kval
from lib.utils import safe_mkdir, safe_rmdir, logger
from lib.filesem import FileSemaphore
from lib.filelock import FileLock, FileLockAcquired
from lib.sharedfile import SharedFile, SharedFileNotFound

__version__ = "0.2.0.1"

class FileCacheException(Exception):
    pass

class FileCacheNotRegularFile(FileCacheException):
    pass


class FileCacheEntryMetadata(dict):
    """ Provies a dictionary interface for metadata """
    def __init__(self, manager, *args, **kwargs):
        super(FileCacheEntryMetadata, self).__init__(*args, **kwargs)
        self.manager = manager

    def __getitem__(self, key):
        f = SharedFile(self.manager.entry.metadata_path, mode="r")
        try:
            f.open()
        except SharedFileNotFound:
            raise KeyError
        try:
            self.update(self.manager.metadata_serializer.load(f))
            return super(FileCacheEntryMetadata, self).__getitem__(key)
        except:
            f.request_rollback()
            raise
        finally:
            f.close()

        pass

    def __setitem__(self, key, value):
        f = SharedFile(self.manager.entry.metadata_path, mode="w")
        f.open()
        try:
            super(FileCacheEntryMetadata, self).__setitem__(key, value)
            self.manager.metadata_serializer.dump(self, f)
        except:
            f.request_rollback()
            raise
        finally:
            f.close()

        logger.debug("metadata key '%s' is now %s" % (key, value))

    def __delitem__(self, key):
        # TODO: metadata file has two dots ..foo.metadata.lock
        f = SharedFile(self.manager.entry.metadata_path, mode="rw")
        try:
            f.open()
        except SharedFileNotFound:
            raise KeyError
        try:
            try:
                self.update(self.manager.metadata_serializer.load(f))
                super(FileCacheEntryMetadata, self).__delitem__(key)
                self.manager.metadata_serializer.dump(self, f)
            except:
                f.request_rollback()
                raise
        finally:
            f.close()


class FileCacheEntryMetadataManager(object):
    """ Provides a means to attach metadata to a cache entry """
    def __init__(self, entry, metadata_serializer=lib.kval):
        """
        entry - pointer back to cache entry
        metadata_serializer - used to read/write metadata to file
        """
        self.entry = entry
        self.metadata_serializer = metadata_serializer

    def set(self, metadata):
        if not metadata:
            return

        f = SharedFile(self.entry.metadata_path, mode="w")
        f.open()
        try:
            self.metadata_serializer.dump(metadata, f)
        except:
            f.request_rollback()
            raise
        finally:
            f.close()

        logger.debug("'%s' metadata is now %s" %
                      (self.entry.key, pformat(metadata)))

    def get(self):
        f = SharedFile(self.entry.metadata_path, mode="r")
        try:
            f.open()
        except SharedFileNotFound:
            return FileCacheEntryMetadata(self)

        try:
            return FileCacheEntryMetadata(
                self, self.metadata_serializer.load(f))
        finally:
            f.close()

    def delete(self, wait=True):
        logger.debug("deleting metadata for '%s'" % self.entry.key)
        try:
            SharedFile(self.entry.metadata_path, mode="d").delete(wait=wait)
        except SharedFileNotFound:
            pass

class FileCacheEntry(SharedFile):
    """ Represents a cached file
   
    Metadata is stored in the .metadata file in a the format defined by the
    metadata_serializer, the default being kval ("a=b").
    """

    def __init__(self, cache, key, mode="r", overwrite=True, wait=True,
                 timeout=10, delay=0.5):
        """ create a new cache entry

        cache - points back to parent cache
        key - key to use, must able to be a filename on a POSIX system
        """
        self.cache = cache
        self.key = key
        path = os.path.join(self.cache.path, self.key)

        super(FileCacheEntry, self).__init__(path, mode=mode,
            overwrite=overwrite, wait=wait, timeout=timeout, delay=delay)

        self.metadata_path = os.path.join(self.cache.path,
                                          ".%s.metadata" % self.key)
        self.metadata_manager = FileCacheEntryMetadataManager(self)

    def delete(self, wait=True):
        del self.metadata
        SharedFile.delete(self, wait=wait)

    def _get_metadata(self):
        return self.metadata_manager.get()
    def _set_metadata(self, metadata):
        self.metadata_manager.set(metadata)
    def _del_metadata(self):
        self.metadata_manager.delete()

    metadata = property(_get_metadata, _set_metadata, _del_metadata)


class FileCache(object):
    """ Provides a file cache entirely implemented using the filesystem.

    N.B.: Keys cannot start with a '.' (dot).
    """
    def __init__(self, path, fetch_fn=None, cache_nx=False, pruning_interval=1):
        """ creates a new FileCache

        path - directory to hold cache
        fetch_fn - function(entry, chunksize) to run on cache miss
        cache_nx - if True, will cache the absense of a record
        pruning_interval - number of cache misses between pruning
        """
        self.path = os.path.abspath(path)
        self.fetch_fn = fetch_fn or getattr(self, "_fetch")
        self.cache_nx = cache_nx
        self.pruning_interval = pruning_interval
        self.deferred_prune_count = 0
        safe_mkdir(self.path)
        self.write_lock = self._make_write_lock()
        self.writers_sem = self._make_writers_semaphore()
        self.delete_lock = self._make_delete_lock()
        self.deleters_sem = self._make_deleters_semaphore()
        logger.debug("(*)FileCache created at '%s'" % self.path)

    def get(self, key, wait=True, chunksize=4096, metadata=None,
            no_fetch=False):
        """ retrieve a given key from the cache

        key - key to retrieve from cache
        wait - true will block if lock() fails, false will raise
        chunksize - size in bytes of chunks to yield
        metadata - optional metadata to set on a fetch (cache miss)
        no_fetch - if True raise not found rather than fetch from origin
        """
        metadata = metadata or {}

        try:
            for chunk in self._emit_from_cache(key, chunksize):
                yield chunk
        except SharedFileNotFound:
            # cache miss
            if no_fetch:
                raise KeyError("key '%s' not found" % key)

            for chunk in self._fetch_and_cache(key, chunksize, metadata, wait):
                yield chunk

            self.prune()

    __getitem__ = get

    def prune(self, wait=True, force=False):
        if self.deferred_prune_count >= self.pruning_interval or force:
            try:
                self.delete_lock.acquire(wait=False)
            except FileLockAcquired:
                # someone else is pruning, move along
                return

            try:
                # allow any deleters to finish, writers can continue
                self.deleters_sem.wait(wait=wait)

                self._prune()
                self.deferred_prune_count = 0
            finally:
                self.delete_lock.release()
        else:
            self.deferred_prune_count += 1
            logger.debug("deferred_prune_count is now %i" %
                           self.deferred_prune_count)

    def delete(self, key, wait=True):
        """ delete a given key from the cache

        key - key to delete from cache
        wait - true will block if lock() fails, false will raise
        """
        start_time = time.time()
        logger.debug("deleting key '%s'" % key)
        try:
            self.delete_lock.acquire()
            try:
                self.deleters_sem.up()
            finally:
                self.delete_lock.release()

            try:
                self._unsafe_delete(key)
            finally:
                self.deleters_sem.down()

        except SharedFileNotFound:
            raise KeyError("key '%s' not found" % key)

        logger.info("'%s' deleted from cache in %.2f secs" % 
                     (key, (time.time() - start_time)))

    __delitem__ = delete

    def clear(self, wait=True, rmdir=False):
        """ clear cache by deleting all keys
       
        wait - true will block if lock() fails, false will raise
        rmdir - if true the cache directory will be removed
        """
        start_time = time.time()
        
        self.write_lock.acquire(wait=wait)
        try:
            self.delete_lock.acquire(wait=wait)
            try:
                # wait on any existing deleters
                self.deleters_sem.wait(wait=wait)

                # wait on any existing writers
                self.writers_sem.wait(wait=wait)
            
                num_keys = 0
                # need to use _unsafe_keys because we want to clear
                # invalidated keys but we already have acquired a delete lock
                for key in self._unsafe_keys():
                    # since the delete lock has been acquired, we need to call
                    # the lock-less delete
                    self._unsafe_delete(key)
                    num_keys += 1

                if rmdir:
                    self._rmdir_path()
            finally:
                self.delete_lock.release()
        finally:
            self.write_lock.release()

        logger.info("cleared %s keys in %.2f secs" % 
                     (num_keys, (time.time() - start_time)))

    def metadata(self, key):
        return FileCacheEntry(self, key).metadata

    def has_key(self, key):
        return key in self.keys()

    __contains__ = has_key

    def len(self):
        return len(self.keys())

    __len__ = len

    # The three key methods merits an explanation:
    #  * keys() is what clients should use; it only return keys that are valid
    #
    #  * _keys() is just like keys() but without the validity check. This
    #    is used for clearing the cache
    #
    #  * unsafe_keys() is used when a method already has acquired a
    #    delete_lock()
    def keys(self):
        return [k for k in self._keys() if self.check_valid(k)]

    def _keys(self):
        return self._stat().keys()

    def _unsafe_keys(self):
        return self._unsafe_stat().keys()

    @property
    def size(self):
        """ the total size of the keys contents in bytes, does not include
        size of metadata
        """
        return sum(s["size"] for s in self._stat().values())

    def check_valid(self, key):
        if '__nx__' in self.metadata(key):
            return False
        return self._check_valid(key)

    def _rmdir_path(self):
        safe_rmdir(self.path)

    def _stat(self):
        self.delete_lock.acquire()
        try:
            return self._unsafe_stat()
        finally:
            self.delete_lock.release()

    def _unsafe_stat(self):
        try:
            fnames = os.listdir(self.path)
        except OSError, e:
            # the cache dir may have been clear(rmdir=True) out from under us,
            # in this case just indicate that the cache has no keys. The cache
            # will be recreated upon next write
            if e.errno != errno.ENOENT:
                raise
            return {}

        cache_stats = {}
        for key in fnames:
            if key.startswith('.'):
                continue
            try:
                cache_stats[key] = self._key_stat(key)
            except FileCacheNotRegularFile:
                continue

        return cache_stats

    def _key_stat(self, key):
        file_info = os.stat(os.path.join(self.path, key))
        mode = file_info[stat.ST_MODE]

        if not stat.S_ISREG(mode):
            raise FileCacheNotRegularFile

        return {
            "ctime": file_info[stat.ST_CTIME],
            "atime": file_info[stat.ST_ATIME],
            "mtime": file_info[stat.ST_MTIME],
            "size": file_info[stat.ST_SIZE]
        }


    def _emit_from_cache(self, key, chunksize):
        start_time = time.time()

        entry = FileCacheEntry(self, key, mode="r")

        if '__nx__' in entry.metadata:
            raise KeyError("key '%s' not found [cached_nx]" % key)

        entry.open()
        try:
            if not self.check_valid(key):
                logger.debug("forced cache miss for '%s'" % key)
                raise SharedFileNotFound

            logger.debug("cache hit for '%s'" % key)
            chunk_reader = lambda: entry.read(chunksize)
            chunk = chunk_reader()
            while chunk:
                yield chunk
                chunk = chunk_reader()
        finally:
            entry.close()

        logger.info("'%s' emitted from cache in %.2f secs" %
                     (entry.key, (time.time() - start_time)))

    def _fetch_and_cache(self, key, chunksize, metadata, wait):
        self.write_lock.acquire(wait=wait)
        try:
            self.writers_sem.up()
            # need to create the path in case another client has
            # clear(rmdir=True) out from under us
            # FIXME: I don't think this is needed. sharedfile should create
            # the directory at the last possible sec before renaming the
            # tmpfile
            safe_mkdir(self.path)
        finally:
            self.write_lock.release()
        try:
            entry = FileCacheEntry(self, key, mode="w")
            entry.open()
            try:
                for chunk in self._unsafe_fetch_and_cache(
                            entry, chunksize, metadata):
                    yield chunk
            finally:
                entry.close()
        finally:
            self.writers_sem.down()

    def _unsafe_fetch_and_cache(self, entry, chunksize, metadata):
        start_time = time.time()
        logger.debug("cache miss for '%s'" % entry.key)

        # FIXME: Writing to the metadata file to just to pass data to the
        # fetcher seems messy. We should probably make metadata an optional
        # argument to the fetch_routine; then we persist the metadata only
        # if we don't raise an exception
        entry.metadata = metadata
        try:
            for chunk in self.fetch_fn(entry, chunksize):
                entry.write(chunk)
                yield chunk

        except KeyError:
            # TODO: should the fetcher raise a custom Exception object rather
            # than just use KeyError?
            if self.cache_nx:
                self._do_cache_nx(entry, metadata)
            else:
                entry.request_rollback()
            raise
        except:
            entry.request_rollback()
            raise

        logger.info("'%s' fetched from remote in %.2f secs" %
                     (entry.key, (time.time() - start_time)))

    def _do_cache_nx(self, entry, metadata):
        metadata['__nx__'] = '1'
        entry.metadata = metadata
        logger.debug("cached nx entry for key '%s'" % entry.key)

    def _unsafe_delete(self, key):
        FileCacheEntry(self, key, mode="d").delete()

    def _make_cache_lock_path(self, suffix):
        # All users of the cache need to agree upon the names of the global
        # locks
        cache_id = md5.new(self.path).hexdigest()
        return os.path.join(gettempdir(), "%s.cache.%s" % (cache_id, suffix))

    def _make_write_lock(self):
        # N.B. to avoid races surrounding clear(), the cache-wide mutex must
        # be stored in a directory outside of the cache.
        return FileLock(self._make_cache_lock_path('write_lock'))

    def _make_delete_lock(self):
        # N.B. to avoid races surrounding clear(), the cache-wide mutex must
        # be stored in a directory outside of the cache.
        return FileLock(self._make_cache_lock_path('delete_lock'))

    def _make_writers_semaphore(self):
        # N.B. needed to count writers so clear() can wait for them to finish
        return FileSemaphore(self._make_cache_lock_path('writers_sem'))

    def _make_deleters_semaphore(self):
        # N.B. needed to count deleters so clear() can wait for them to finish
        return FileSemaphore(self._make_cache_lock_path('deleters_sem'))

    def _prune(self):
        """ Subclasses can override this to provide a pruning strategy """
        pass

    def _check_valid(self, key):
        """ Subclasses can override this to force a cache miss. This can be
        used to implement expiry.
        """
        return True

class LRUFileCache(FileCache):
    """
    Provides a least-recently-used file cache.

    Tune keys_pct and size_pct to avoid thrashing near the cut-off point.

    Use pruning_interval to avoid having to the prune the cache on every cache
    miss. Note that anything value other than one may cause the cache to
    exceed the stated max_keys or max_size. Take this into account.
    """
    def __init__(self, path, fetch_fn=None,
                 max_keys=None, keys_pct=0.75,
                 max_size=None, size_pct=0.75, 
                 prune_dry_run=False, pruning_interval=10):
        """
        path - directory to hold cache
        fetch_fn - function(entry, chunksize) for fetching item on cache miss

        max_keys - maximum number of keys before pruning
        keys_pct - begin pruning at keys_pct of max_keys (hysteresis to avoid
                   thrashing around limit)

        max_size - maximum size of cache in bytes
        size_pct - begin pruning at size_pct of max_size

        prune_dry_run - log pruned files, but don't delete them
        pruning_interval - number of gets between pruning
        """
        if max_keys is None and max_size is None:
            raise FileCacheException(
                "must specify max_keys, max_size, or both")

        super(LRUFileCache, self).__init__(path, fetch_fn=fetch_fn,
                                           pruning_interval=pruning_interval)
        
        self.max_keys = max_keys
        self.keys_pct = keys_pct

        if max_keys is not None:
            logger.debug("key constraint set to %i max with %.2f full" %
                         (max_keys, keys_pct))

        self.max_size = max_size
        self.size_pct = size_pct

        if max_size is not None:
            logger.debug("size constraint set to %i max with %.2f full" %
                         (max_size, size_pct))

        self.prune_dry_run = prune_dry_run

    def _prune(self):
        start_time = time.time()
        num_keys = 0
        # N.B. prune already as the delete lock
        cache_stats = self._unsafe_stat()
        logger.debug("cache_stats for pruning %s" % pformat(cache_stats))

        lru_keys = self._make_lru_keys(cache_stats)
        num_keys += self._prune_max_keys(cache_stats, lru_keys)
        num_keys += self._prune_max_size(cache_stats, lru_keys)

        logger.info("pruned %s keys in %.2f secs" % 
                     (num_keys, (time.time() - start_time)))

    def _make_lru_keys(self, cache_stats):
        """ sort keys newest (highest) -> oldest (lowest) """
        return [
            k for k, v in sorted(
            cache_stats.items(), key=lambda (k, v): (v['atime'], v['mtime']),
            reverse=True)
        ]

    def _prune_max_keys(self, cache_stats, lru_keys):
        if self.max_keys is None:
            return 0

        num_keys = len(lru_keys)
        ok_keys = ceil(self.keys_pct * self.max_keys)
        to_free = num_keys - ok_keys
        if to_free <= 0:
            return 0

        freed = 0
        logger.debug("%i/%i (max: %i) freeing %i items" %
                     (num_keys, ok_keys, self.max_keys, to_free))

        while lru_keys and to_free > 0:
            key = lru_keys.pop()

            if self.prune_dry_run:
                logger.debug("would have pruned '%s'" % key)
            else:
                # since the delete lock has been acquired, we need to call
                # the lock-less delete
                self._unsafe_delete(key)
            
            to_free -= 1
            freed += 1

        return freed

    def _prune_max_size(self, cache_stats, lru_keys):
        if self.max_size is None:
            return 0

        cache_size = sum(v['size'] for v in cache_stats.values())
        ok_size = ceil(self.size_pct * self.max_size)
        to_free = cache_size - ok_size

        if to_free <= 0:
            return 0

        freed = 0
        logger.debug("%ib/%ib (max: %ib) trying to free %i bytes" %
                     (cache_size, ok_size, self.max_size, to_free))

        while lru_keys and to_free > 0:
            key = lru_keys.pop()
            del_size = cache_stats[key]['size']

            if self.prune_dry_run:
                logger.debug("would have pruned '%s' for %i bytes" 
                              % (key, del_size))
            else:
                # since the delete lock has been acquired, we need to call
                # the lock-less delete
                self._unsafe_delete(key)

            to_free -= del_size
            freed += 1

        return freed

class ExpiryFileCache(FileCache):
    """
    Provides an expiry file cache
    """
    def __init__(self, path, fetch_fn=None, default_expiry=None,
                 prune_dry_run=False, pruning_interval=10):
        super(ExpiryFileCache, self).__init__(
            path, fetch_fn=fetch_fn, pruning_interval=pruning_interval)
        self.default_expiry = default_expiry
        self.prune_dry_run = prune_dry_run

    def _check_valid(self, key):
        return not self._check_stale(key)

    def _check_stale(self, key):
        try:
            expiry = float(self.metadata(key)['expiry'])
        except ValueError:
            # bad value just skip it
            # TODO: should we raise? not sure
            expiry = None
        except KeyError:
            expiry = None

        expiry = expiry or self.default_expiry

        if expiry is None:
            return True
        
        age = time.time() - self._key_stat(key)["mtime"]
        logger.debug("key '%s' age %.2f secs with expiry %.2f"
                     % (key, age, expiry))
        return age > expiry

    def _prune(self):
        start_time = time.time()
        num_keys = 0
        # N.B. prune already as the delete lock
        cache_stats = self._unsafe_stat()
        logger.debug("cache_stats for pruning %s" % pformat(cache_stats))
        for key in cache_stats:
            if self._check_stale(key):
                self._unsafe_delete(key)
                num_keys += 1

        logger.info("pruned %s keys in %.2f secs" % 
                     (num_keys, (time.time() - start_time)))

