"""
Provides a file-based counting semaphore
"""
import os
import time
from filelock import FileLock

class FileSemaphoreException(Exception):
    pass

class FileSemaphoreOverflow(FileSemaphoreException):
    pass

class FileSemaphoreUnderflow(FileSemaphoreException):
    pass

class FileSemaphoreTimeout(FileSemaphoreException):
    pass

class FileSemaphoreInUse(FileSemaphoreException):
    pass

class FileSemaphore(object):
    def __init__(self, sem_path, min_value=None, max_value=None,
                 timeout=10, delay=0.5):
        self.sem_path = sem_path
        self.max_value = max_value
        self.min_value = min_value
        self.timeout = timeout
        self.delay = delay
        self.lock = FileLock("%s.lock" % sem_path, timeout=timeout, delay=delay)

    def up(self, n=1):
        return self._shift(n)

    def down(self, n=1):
        return self._shift(-n)

    def wait(self, n=0, wait=True):
        """ Wait for semaphore to take on a particular value """
        start_time = time.time()
        while True:
            if self._shift(0) == n:
                break
            if not wait:
                raise FileSemaphoreInUse("'%s' in use" % self.sem_path)
            if (time.time() - start_time) >= self.timeout:
                raise FileSemaphoreTimeout("timeout for '%s'" % self.sem_path)
            time.sleep(self.delay)

    def __enter__(self):
        self.up()
        return self
 
    def __exit__(self, type, value, traceback):
        self.down()

    def _shift(self, n):
        self.lock.acquire()
        try:
            return self._unsafe_shift(n)
        finally:
            self.lock.release()

    def _unsafe_shift(self, n):
        new_value = None
        self.fd = os.open(self.sem_path, os.O_CREAT|os.O_RDWR)
        try:
            buf = os.read(self.fd, 16)

            if buf:
                try:
                   value = int(buf) 
                except ValueError:
                    raise FileSemaphoreException("corrupt semaphore file")
            else:
                value = 0

            new_value = value + n
            
            if self.max_value is not None and new_value > self.max_value:
                raise FileSemaphoreOverflow(
                    "value exceeds %i" % self.max_value)
            elif self.min_value is not None and new_value < self.min_value:
                raise FileSemaphoreUnderflow(
                    "value does not exceed %i" % self.min_value)

            os.lseek(self.fd, 0, 0)
            os.write(self.fd, str(new_value))
            return new_value
        finally:
            os.close(self.fd) # should cause an fsync()
            if new_value is not None and new_value == 0:
                os.unlink(self.sem_path)

