import os
import tempfile
from filelock import FileLock
from filesem import FileSemaphore
from utils import logger

class SharedFileException(Exception):
    pass

class SharedFileNotFound(SharedFileException):
    pass

class SharedFileRollback(SharedFileException):
    """ Triggers a rollback from within a 'with' block """
    pass


class SharedFile(object):
    """ Provides shared access to a file.
    
    The atomicity of writes is guaranteed by writing to a tempfile and then
    using an atomic rename.
    
    The semaphore prevents deletion while other readers have the file open.

    By default, SharedFile uses file-based locks and semaphore; however, a
    subclass can use any lock and semaphore that provides the acquire/release,
    up/down semantics.
    """
    def __init__(self, path, mode="r", overwrite=True, wait=True,
                 timeout=10, delay=0.5):
        self.path = path
        self.mode = mode
        self.overwrite = overwrite
        self.wait = wait
        self.timeout = timeout
        self.delay = delay
        self.needs_rollback = False
        self.closed = True
        self.finalized = True
        self.lock = self._make_lock()
        self.readers_sem = self._make_readers_semaphore()

    def open(self, wait=True):
        if self._is_reader_writer():
            self.lock.acquire(wait=wait)
            try:
                self._unsafe_open_read()
                self._open_write()
            except:
                self.lock.release()
                raise
        elif self._is_reader():
            self._open_read(wait=wait)
        elif self._is_writer():
            self._open_write()
        elif self._is_deleter():
            raise SharedFileException("cannot open in delete mode")
        else:
            raise SharedFileException("invalid mode")

        self.closed = False

    def close(self):
        if self._is_reader_writer():
            try:
                self._unsafe_close_read()
                self._close_write()
            except:
                self.lock.release()
                raise
        elif self._is_reader():
            self._close_read()
        elif self._is_writer():
            self._close_write()
        elif self._is_deleter():
            raise SharedFileException("cannot close in delete mode")
        else:
            raise SharedFileException("invalid mode")

        self.closed = True

    def write(self, data):
        if self.closed:
            raise SharedFileException(
                "'%s' is not open, cannot write" % self.path)

        if not self._is_writer():
            raise SharedFileException("must be in write mode")

        if not data:
            return
        
        try:
            return os.write(self.tmp_fd, data)
        except:
            self.request_rollback()
            raise

    def read(self, *args, **kwargs):
        self._read_guard()
        return self.file.read(*args, **kwargs)

    def readlines(self, *args, **kwargs):
        self._read_guard()
        return self.file.readlines(*args, **kwargs)

    def __iter__(self):
        self._read_guard()
        return self.file.__iter__()

    def delete(self, wait=True):
        if not self.closed:
            raise SharedFileException(
                "'%s' is open, cannot delete" % self.path)
      
        if not self._is_deleter():
            raise SharedFileException("must be in delete mode")

        self.lock.acquire(wait)
        try:
            if not os.path.exists(self.path):
                raise SharedFileNotFound("'%s' not found" % self.path)
            
            # wait for any readers to finish
            self.readers_sem.wait(wait=wait)
            
            os.unlink(self.path)
        finally:
            self.lock.release()

    def request_rollback(self):
        """ Signal that we should rollback on close() """
        self.needs_rollback = True

    def commit(self, wait=True):
        if self.closed:
            raise SharedFileException(
                "'%s' is not open, nothing to commit" % self.path)

        if self.needs_rollback:
            raise SharedFileException('rollback needed, cannot commit')

        if self.overwrite:
            self._commit()
        else:
            self.lock.acquire(wait)
            try:
                if os.path.exists(self.path):
                    self.rollback()
                else:
                    # TODO: we made need to safe_mkdir here in case another
                    # user has rmdir'd our directory out from under us
                    self._commit()
            finally:
                self.lock.release()

        self.finalized = True
        logger.debug("commited sharedfile '%s'" % self.path)

    def rollback(self):
        if self.closed:
            raise SharedFileException(
                "'%s' is not open, nothing to rollback" % self.path)

        os.unlink(self.tmp_path)
        self.needs_rollback = False
        self.finalized = True

        logger.debug("rolledback sharedfile '%s'" % self.path)

    def _open_read(self, wait=True):
        self.lock.acquire(wait)
        try:
            if os.path.exists(self.path):
                self.readers_sem.up()
            else:
                raise SharedFileNotFound("'%s' not found" % self.path)
        finally:
            self.lock.release()

        self._unsafe_open_read()

    def _unsafe_open_read(self):
        self.file = open(self.path, self.mode)

    def _close_read(self):
        if not self.closed:
            try:
                self.readers_sem.down()
            finally:
                self._unsafe_close_read()

    def _unsafe_close_read(self):
        self.file.close()

    def _open_write(self):
        self.tmp_fd, self.tmp_path = tempfile.mkstemp()
        self.finalized = False

    def _close_write(self):
        if self.finalized:
            return

        if self.needs_rollback:
            self.rollback()
        else:
            self.commit()

    def _is_reader_writer(self):
        return self.mode == "rw"

    def _is_reader(self):
        return self.mode in ('r', 'rb', "rw")

    def _is_writer(self):
        return self.mode in ('w', 'wb', "rw")

    def _is_deleter(self):
        return self.mode == 'd'

    def _read_guard(self):
        if self.closed:
            raise SharedFileException(
                "'%s' is not open, cannot read" % self.path)

        if not self._is_reader():
            raise SharedFileException("must be in read mode, not '%s'" %
                                      self.mode)

    def _commit(self):
        # the parent directory may have been deleted out from under us; if
        # that has happened we probably can't avoid a race condition
        os.rename(self.tmp_path, self.path)

    def __enter__(self):
        self.open()
        return self
 
    def __exit__(self, type, value, traceback):
        if isinstance(value, SharedFileRollback):
            self.request_rollback()

        self.close()
 
    def __del__(self):
        if not self._is_deleter:
            self.close()

    def _make_hidden_file_with_suffix(self, suffix):
        path, fname = os.path.split(self.path)
        if not fname.startswith('.'):
            fname = ".%s" % fname
        return os.path.join(path, "%s.%s" % (fname, suffix))

    def _make_lock(self):
        """ Subclasses can override this to use a different locking mechanism
        """
        return FileLock(self._make_hidden_file_with_suffix('lock'))

    def _make_readers_semaphore(self):
        """ Subclasses can override this to use a different semaphore mechanism
        """
        return FileSemaphore(self._make_hidden_file_with_suffix('readers_sem'))

