import time
from filecache import (FileCache, LRUFileCache, ExpiryFileCache,
                       FileCacheException)
from common import URLFileCache, TEST_URLS

def url_test(url):
    cache = URLFileCache('/tmp/url_test_cache', max_keys=100,
                         pruning_interval=1)
    try:
        assert url not in cache
        
        for chunk in cache[url]:
            pass
        
        assert url in cache

        for chunk in cache[url]:
            pass

        assert len(cache) == 1

        del cache[url]
        
        assert url not in cache
        assert cache.size == 0
    finally:
        cache.clear(rmdir=True)

def expiry_test():
    def dummy_fetcher(entry, chunksize):
        key = entry.key
        if key == "foo":
            yield "bar"
        elif key == "spam":
            yield "eggs"
            entry.metadata = { 'expiry': '2.0' }
        elif key == "the day":
            yield "is shot"
        else:
            raise KeyError

    cache = ExpiryFileCache('/tmp/expiry_test_cache', dummy_fetcher,
                            default_expiry=1.0, pruning_interval=1)

    try:
        for chunk in cache['foo']:
            pass
        assert 'foo' in cache
        time.sleep(1.1)
        assert 'foo' not in cache
        # generate a cache miss to trigger a purge
        for chunk in cache['spam']:
            pass
        assert 'foo' not in cache._keys()

        for chunk in cache['foo']:
            pass

        # FIXME:
        # This sleep 1.1 should be sufficient, though interestingly enough, I
        # ran across a failure. Seems the mtime was later than it should have
        # been. Perhaps async fs processing ?
        time.sleep(2.1)

        try:
            for chunk in cache.get('foo', no_fetch=True):
                pass
        except KeyError:
            pass
        else:
            assert False

        for chunk in cache['spam']:
            pass

        assert 'spam' in cache
        time.sleep(1.1)
        # FIXME: Some async process is causing this to randomly fail
        assert 'spam' in cache
        time.sleep(1.1)
        assert 'spam' not in cache

    finally:
        cache.clear(rmdir=True)

def basic_test():
    def dummy_fetcher(entry, chunksize):
        key = entry.key
        if key == "empty":
            yield ""
        else:
            raise KeyError

    cache = FileCache('/tmp/basic_test_cache', dummy_fetcher)
    try:
        # even empty records should be cached
        assert 'empty' not in cache
        for chunk in cache['empty']:
            pass
        assert 'empty' in cache


    finally:
        cache.clear(rmdir=True)


def nx_test():

    def dummy_fetcher(entry, chunksize):
        key = entry.key
        if key == "foo":
            yield "bar"
        else:
            raise KeyError

    cache = FileCache('/tmp/nx_test_cache', dummy_fetcher, cache_nx=True)
    try:
        assert 'foo' not in cache
        for chunk in cache['foo']:
            pass
        assert 'foo' in cache

        # nx records should be cached as well 
        assert 'spam' not in cache
        try:
            for chunk in cache['spam']:
                pass
        except KeyError:
            pass
        else:
            assert False

        assert 'spam' not in cache

        # The KeyError should have been from a cached_nx record
        try:
            for chunk in cache['spam']:
                pass
        except KeyError, e:
            assert 'cached_nx' in str(e)
        else:
            assert False
    finally:
        cache.clear(rmdir=True)


def lru_test():
    def dummy_fetcher(entry, chunksize):
        key = entry.key
        if key == "foo":
            yield "bar"
        elif key == "spam":
            yield "eggs"
        elif key == "the day":
            yield "is shot"
        elif key == "empty":
            yield ""
        else:
            raise KeyError

    try:
        cache = LRUFileCache(None, dummy_fetcher)
    except FileCacheException:
        pass
    else:
        assert False, "should have raised"

    cache = LRUFileCache('/tmp/lru_test_cache', dummy_fetcher, max_keys=2,
                         keys_pct=1.0, pruning_interval=2)
    try:
        assert 'foo' not in cache
        for chunk in cache['foo']:
            pass
        time.sleep(1.01)
        assert 'foo' in cache

        assert 'spam' not in cache
        for chunk in cache['spam']:
            pass
        time.sleep(1.01)
        assert 'spam' in cache

        assert 'the day' not in cache
        for chunk in cache['the day']:
            pass
        time.sleep(1.01)
        assert 'the day' in cache

        # foo should have been pruned
        assert 'foo' not in cache

        del cache['spam']
        del cache['the day']
    finally:
        cache.clear(rmdir=True)

def metadata_test():
    def dummy_fetcher(entry, chunksize):
        key = entry.key
        if key == "foo":
            yield "bar"
            entry.metadata = { 'baz': 'bam', 'lorum': 'ipsum' }
        elif key == "spam":
            yield "eggs"
        elif key == "the day":
            yield "is shot"
        else:
            raise KeyError
    
    cache = FileCache('/tmp/metadata_test_cache', dummy_fetcher)
    try:
        for chunk in cache['foo']:
            pass

        for chunk in cache['spam']:
            pass

        assert 'bam' == cache.metadata('foo')['baz']
        assert 'ipsum' == cache.metadata('foo')['lorum']

        del cache.metadata('foo')['lorum']

        assert 'baz' in cache.metadata('foo')
        assert 'lorum' not in cache.metadata('foo')

        cache.metadata('foo')['the'] = 'game'

        assert 'game' == cache.metadata('foo')['the']

        assert {} == cache.metadata('spam')

        del cache['foo']
        del cache['spam']
    finally:
        cache.clear(rmdir=True)


if __name__ == "__main__":
    nx_test()
    basic_test()
    expiry_test()
    lru_test()
    metadata_test()
    for url in TEST_URLS:
        url_test(url)


